package dmc;

/* ChatInputHandler.java
 * Matthew Wronka
 *
 * Handles input received from JChatPanel's input box.
 * Copyright (C) 2002 Matthew Wronka <wronkm@rpi.edu>
 * This file is released under the GNU General Public License version 2,
 * a version of which should have been provided along with this file.
 *
 * You might also find a copy at:
 * http://dmc.sf.net/files/gpl.txt
 * http://www.fsf.org/copyleft/gpl.html
 */


import javax.swing.*;
import javax.swing.text.*;

import java.awt.Rectangle;

import java.net.UnknownHostException;
import java.io.IOException;

import java.util.StringTokenizer;

import wtalk.WTalk;

/**
 * Handles input received from JChatPanel's input box.
 * Also handles encryption/decryption requests on the client side,
 * and negotiation of encrypted channels.
 *
 * @author Matthew Wronka <wronkm@rpi.edu>
 * @version 0
 */
class ChatInputHandler {
	private static final int RSA_KEY_LENGTH = 2048;
	private static final RSA rsa = new RSA(RSA_KEY_LENGTH);
	private static final int CRYPTO_NULL = 0;
	private static final int CRYPTO_BLOWFISH= 1;
	private int cryptoAlgo = CRYPTO_NULL;
	private String cryptoName = "null";

	private static final java.security.SecureRandom random = new java.security.SecureRandom();
	private TabbedChat tc = null;
	private JTextPane tp = null;
	private WTalk wt = null;
	private String CurrentRoom = null;
	private Document doc = null;

	private Object Encryptor=null;
	private String partnersKey=null;
	private byte[] sharedKey = null;

	/**
	 * Initializes internal variables
	 *
	 * @param room the room or other member of a conversation represented
	 * 	in the parent JChatPanel.
	 * @param t The text pane in the JChatPanel.
	 * @param c TabbedChat object which is the parent 
	 *	of our parent JChatPanel.
	 * @param w the WTalk object used by TabbedChat c.
	 */
	public ChatInputHandler(String room, JTextPane t, TabbedChat c, WTalk w) {
		CurrentRoom = room;
		tp = t;	wt = w; tc = c;
		doc = t.getDocument();
	}

	/**
	 * Prints a string parent JChatPanel.
	 *
	 * @param s String to write to the JChatPanel.
	 */
	public void printChat(String s) {
		try {
			doc.insertString(doc.getLength(), s, tp.getStyle("client"));
			tp.scrollRectToVisible(new Rectangle(0, tp.getHeight(), 1, 1));
		} catch (BadLocationException ble) {}
	}

	/**
	 * Parses the input from JChatPanel and performs appropriately,
	 * 	probably by calling <code>doCommand()</code>, 
	 *	or by printing an error.
	 *
	 * @s Input line to parse.
	 */
	public void doInput(String s) {
		if(s.length() <= 0) return;

		if(s.charAt(0)=='/') {
			if(s.length() >= 2 && s.charAt(1)=='/') 
				s=s.substring(1);
			else{
				doCommand(s.substring(1));
				return;
			}
		}

		if(CurrentRoom != null && s!=null && s.length() > 0)
			doCommand("msg " + CurrentRoom + " " + encrypt(s));
		else printChat(ErrorMessages.NO_ACTIVE_ROOM);
	}

	/**
	 * Executes the internal command contained in <em>s</em>.
	 * <em>s</em> contains the command and arguments, but no
 	 * initial forward slash, '/'.
	 *
	 * @s Command string to execute. 
	 */
	public void doCommand(String s) {
		StringTokenizer st = new StringTokenizer(s);
		String com;

		if(st.hasMoreTokens()){
			com = st.nextToken();


			//
			// UI Commands
			//
			if(com.equals("clear")) {
				try {
					doc.remove(0, doc.getLength());
				} catch ( BadLocationException ble) {}
				return;
			} else if (com.equals("close")) {
				String room;
				if(st.hasMoreTokens()) {
					room = st.nextToken();
				} else {
					room = CurrentRoom;
				}

				if(room!=null) {
					int i = tc.getUI().indexOfTab(room);
					if(i>=0) tc.getUI().removeTabAt(i);
				}
				return;	
			} else if (com.equals("open")) {
				if(st.hasMoreTokens()) {
					String room = st.nextToken();
					if(room!=null) {
						int i = tc.getUI().indexOfTab(room);
						if(i<0) tc.addTab(room, JChatPanel.IM_TAB);
					}
				} 
				return;
			} else if (com.equals("newconn")) {
				String connName=null, connType=null;
				if(st.hasMoreTokens()) 
					connName = st.nextToken();
				if(st.hasMoreTokens()) 
					connType = st.nextToken();

				if(connType!=null && connName!=null ){
					tc.getTabbedConnection().openConnection(connName, connType);
				} else {
					printChat("Please try '/newconn <connection name> <connection type>'\n");
				}
				return;
			} else if (com.equals("closeconn")){
                                String connName;
                                if(st.hasMoreTokens()) {
                                        connName = st.nextToken();
                                } else {
                                        connName = tc.getTabbedConnection()
						.getUI().getTitleAt(
							tc.getTabbedConnection().getUI().getSelectedIndex()
						);
                                }

				if(tc.getTabbedConnection().getUI()
					.getTabCount()<=1) {
					//Don't let them delete all tabs
					printChat("Sorry, but you can't delete the final connection!\n"); 

					return;
				}

                                if(connName!=null) {
                                        int i = tc.getTabbedConnection().getUI().indexOfTab(connName);
                                        if(i>=0) { 
						((JChatPanel)((JTabbedPane)tc
							.getTabbedConnection()
							.getUI()
							.getComponentAt(i))

							.getSelectedComponent())
							.doCommand(
							"quit DMC: Connection Closed.");
						tc.getTabbedConnection().getUI().removeTabAt(i);
					}
					
                                }
                                return;
			} else if (com.equals("exit")){
				// Should probably clean up at some point...
				System.exit(0);
				return;
			}


			//
			// Connection commands
			//
			if(com.equals("server")) {
				if(st.hasMoreTokens()) {
					int port;
					String server = st.nextToken();
					if(st.hasMoreTokens()){
						port = Integer.parseInt(st.nextToken());
						wt.getWTalkProtocol().setServerPort(port);
					}

					wt.getWTalkProtocol().setServer(server);
				}
				return;
			} else  if(com.equals("connect")) {

					if(wt.getWTalkProtocol().isConnected()){
						doCommand("quit");
						printChat("Disconnected.\n");
					}



					String server = wt.getWTalkProtocol().getServer();
					int port = wt.getWTalkProtocol().getServerPort();
					try {
						printChat("Trying to connect: " + server + ":" + port  + "\n"); 
						if(st.hasMoreTokens())
							wt.getWTalkProtocol().connect(st.nextToken());
						else
							wt.getWTalkProtocol().connect();
					} catch (UnknownHostException e) {
						printChat("Unknown Host: " + server + "\n"); 
					} catch (IOException e) {
						printChat("Error Connecting: " + server + ":" + port + "\n");
					}
				return;
			}


			//
			// Remembered Variables
			//
			if (com.equals("nick")) {
				if(st.hasMoreTokens()) {
					String nick = st.nextToken();
					wt.getWTalkProtocol().setName(nick); 
					return;
				}else{
					printChat( "You are currently known as " + tc.getUserName() + "\n");
					return;
				}
			} else if (com.equals("passwd") || 
					com.equals("password")) {
				if(st.hasMoreTokens()) {
					String passwd = st.nextToken();
					wt.getWTalkProtocol().setPassword(passwd);
				}
				return;
			}



			//
			// Connected Only
			//
			if(wt.getWTalkProtocol() == null || !wt.getWTalkProtocol().isConnected()) {
				printChat(ErrorMessages.NOT_CONNECTED); 
				return;
			}





			if(com.equals("msg")) {
				if(st.hasMoreTokens()) {
					String to = st.nextToken();
					if(st.hasMoreTokens())
						wt.getWTalkProtocol().sendMessage(to, s.substring(to.length()+com.length()+2));
				}
			} else if (com.equals("encrypt")) {
				doCommand("secure " + s.substring(com.length()+1));
				requestPublicKey();
				//doCommand("negotiate");
			} else if (com.equals("secure")) {
				manageSecure(s.substring(com.length()+1));
			} else if (com.equals("negotiate")) {
				negotiateSecure();
			} else if (com.equals("wall")) {
				if(st.hasMoreTokens()) {
					wt.getWTalkProtocol().sendWall(s.substring(com.length()+1));
				}
			} else if (com.equals("sub") || com.equals("ctcp")) {
				String to=null;
				String data=null;
				if(CurrentRoom != null) {
					to=CurrentRoom;
					data = s.substring(com.length()+1);
				} else {
					if(st.hasMoreTokens()) to=st.nextToken();
					data = s.substring(to.length()+com.length()+2);
				}
				if(to!=null) {
					wt.getWTalkProtocol().sendSub(to,data);
				} else {
					// XXX Error Message
				}
			} else if (com.equals("ban")) {
				while(st.hasMoreTokens()) {
					wt.getWTalkProtocol().setBan(st.nextToken(), true);
				}
			} else if (com.equals("unban")) {
				while(st.hasMoreTokens()) {
					wt.getWTalkProtocol().setBan(st.nextToken(), false);
				}
			} else if (com.equals("ignore")) {
				if(st.hasMoreTokens()) {
					String troutWielder = st.nextToken();
					String sIgnoreType = null;

					if(st.hasMoreTokens())
						sIgnoreType = st.nextToken();

					tc.addIgnore(troutWielder, sIgnoreType);

					printChat("Now ignoring " + troutWielder + "\n");
				} else {
					printChat(tc.showIgnored() + "\n");
				}
			} else if (com.equals("unignore")) {
				if(st.hasMoreTokens()) {
					String troutWielder = st.nextToken();
					String sIgnoreType = null;

					if(st.hasMoreTokens())
						sIgnoreType = st.nextToken();

					tc.delIgnore(troutWielder, sIgnoreType);

					printChat("No logner ignoring " + troutWielder + "\n");
				} else {
				}
			} else if (com.equals("unignore")) {
				if(st.hasMoreTokens()) {
					String troutWielder = st.nextToken();
					printChat("No longer ignoring " + troutWielder + "\n");
				} else {
				} 
			} else if (com.equals("quit")) {
				if(st.hasMoreTokens()) {
					wt.getWTalkProtocol().disconnect(s.substring(com.length()+1));
				} else {
					wt.getWTalkProtocol().disconnect(null);
				}
				 printChat(ErrorMessages.DISCONNECTED);
			} else if (com.equals("away")) {
				if(st.hasMoreTokens()) {
					// set away message	
					wt.getWTalkProtocol().setAway(s.substring(com.length()+1));
				} else {
					// remove away (send null away)
					wt.getWTalkProtocol().setAway(null);
				}
			} else if(com.equals("join")) {
				if(st.hasMoreTokens()){
					String room=s.substring(com.length()+1);
					wt.getWTalkProtocol().joinChat(room);
					// If there isn't a tab already ...
					if(tc.getUI().indexOfTab(room) < 0)
						tc.addTab(room, JChatPanel.CHAT_TAB);
				} else if(CurrentRoom != null) {
					wt.getWTalkProtocol().joinChat(CurrentRoom);
				} else {
					printChat(ErrorMessages.NO_ROOM_SPECIFIED);
				}
			} else if(com.equals("leave") || com.equals("part")) {
				if(st.hasMoreTokens()) {
					wt.getWTalkProtocol().partChat(s.substring(com.length()+1));
				} else if(CurrentRoom != null) {
					wt.getWTalkProtocol().partChat(CurrentRoom);
				} else {
					printChat(ErrorMessages.NO_ROOM_SPECIFIED);
				}
			} else if (com.equals("list")) {
				if(st.hasMoreTokens()) {
					wt.getWTalkProtocol().chatList(s.substring(com.length()+1));
				} else {
					wt.getWTalkProtocol().chatList(null);
				}
			} else if (com.equals("addbuddy")) {
				while(st.hasMoreTokens()) {
					handleBuddy(st.nextToken(), true);
				}
			} else if (com.equals("delbuddy")) {
				while(st.hasMoreTokens()) {
					handleBuddy(st.nextToken(),false);
				}
			} else if (com.equals("topic")) {
				String room;	int len=0;

				if(CurrentRoom!=null) {
					room=CurrentRoom;
					len = com.length()+1;
				} else if(st.hasMoreTokens()){
					room=st.nextToken();
					len = room.length()+com.length()+2;
				} else {
					printChat(ErrorMessages.NO_ROOM_SPECIFIED);
					return;
				}
				if(s.length() > len)
					wt.getWTalkProtocol().setTopic(room, s.substring(len));
				else
					wt.getWTalkProtocol().setTopic(room, null);

			} else if (com.equals("whois") || com.equals("wi")) {
				String who =  null;
				if (st.hasMoreTokens()) {
					who = s.substring(com.length() + 1);
				} else if (CurrentRoom != null) {
					who = CurrentRoom;
				}

				if (who!=null) {
					wt.getWTalkProtocol().whois(who);
				} else {
					printChat(ErrorMessages.NO_USER_SPECIFIED);
				}
			} else if (com.equals("send") || com.equals("raw")) {
				wt.getWTalkProtocol().sendRaw(s.substring(com.length() + 1));
			} else {
				 printChat(ErrorMessages.UNKNOWN_CMD);
			}
		}
	}


	/**
	 * Modifies the user's buddy (user) list.
	 * <p>
	 * The String <p>s</p> should be of the form:
	 * <pre>group:user1[,user2[,...]]</pre>
	 *
	 * @param s Group and list of buddies to add/remove.
	 * @param bAdd If true, buddies are added;
	 *	otherwise they are removed.
	 *
	 */
	// bAdd := true, then add buddy, false delete
	private void handleBuddy(String s, boolean bAdd) {
		StringTokenizer stg = new StringTokenizer(s, ":");
		StringTokenizer stu;
		String group ="";
		String s2 ;

		if (stg.countTokens()>=1) {
			group = stg.nextToken();
			s2= s.substring(group.length()+1); 
		} else  s2=s;
		
		stu = new StringTokenizer(s2, ",");
		while(stu.hasMoreTokens()) {
			if(bAdd)
				wt.getWTalkProtocol().addBuddy(stu.nextToken(), group);
			else
				wt.getWTalkProtocol().removeBuddy(stu.nextToken(), group);

		}
	}

	/**
	 * Sets the encryption method used.
	 * <p>
	 * Presently, only "null" and "blowfish" are supported.
	 *
	 * @param s Name of encryption algorithm to use.
	 * @result Returns true if the algorithm is supported.
	 *	Otherwise, false is returned.
	 */
	public boolean manageSecure(String s) {
		boolean rval=false;
		if(s==null || s.equals("null")) {
			cryptoAlgo = CRYPTO_NULL;
			rval=true;
		}else if (s.equals("blowfish")) {
			cryptoAlgo = CRYPTO_BLOWFISH;
			Encryptor = new BlowfishJ.BlowfishEasy(s);
			rval=true;
		}

		cryptoName=s;
		printChat("Now using " + s + " for encryption.\n");
		return rval;
	}

	/**
	 * Encrypts <em>s</em> with the current algorithm.
	 *
	 * @param s Cleartext message.
	 * @result Ciphertext version of <em>s</em>
	 */
	public String encrypt(String s) {
		switch(cryptoAlgo) {
		case CRYPTO_BLOWFISH:
			return ((BlowfishJ.BlowfishEasy)Encryptor).encryptString(s);
		case CRYPTO_NULL:
		default:
			return s; 
		}
	}

        /**
         * Decrypts <em>s</em> with the current algorithm.
         *
         * @param s Ciphertext message.
         * @result Cleartext version of <em>s</em>
         */
	public String decrypt(String s) {
		switch(cryptoAlgo) {
		case CRYPTO_BLOWFISH:
			return ((BlowfishJ.BlowfishEasy)Encryptor).decryptString(s);
		case CRYPTO_NULL:
		default:
			return s;
		}
	}

	/**
	 * Sets the shared key for use in symetric encryption.
	 *
	 * @param ba Shared key to be used. 
	 */
	public void setSharedKey(byte[] ba){
		sharedKey = ba;
	}

	/**
	 * Stores communication partner's public key.
	 * This is used for RSA encryption to pass shared secrets.
	 *
	 * @param key Partner's public RSA key, in the form, "e n".
	 */
	public void setPublicKey(String key){
		partnersKey = key;
	}

	/**
	 * Sends a request for partner's public key.
	 */
	private void requestPublicKey(){
		wt.getWTalkProtocol().sendMessage(CurrentRoom,  "\1<!--Seeing if you'll give me a key...--> DMC-negotiating"  );
	}

	/**
	 * Sends public key to partner.
	 */
	public void sendPublicKey(){
		wt.getWTalkProtocol().sendMessage(CurrentRoom,  "\2dmc-key " + rsa.getPublicKey());
		
	}

	/**
	 * Decrypts the RSA encrypted message <em>msg</em> using 
	 *	our private key.
	 *
	 * @param msg RSA encrypted message.
	 * @return Cleartext version of <em>msg</em>.
	 */
	public String decryptRSA(String msg) {
		return rsa.decrypt(new java.math.BigInteger(msg,16));
	}

	/**
	 * Negotiate's a secure connection with partner.
	 * <p>
	 * If we do not already have their public key, we request it.
	 * If we have their public key, we create a shared key, encrypt
	 *	it with their public key, and send it off.
	 */
	// sends a message with key
	public void negotiateSecure() {
		String msg;

		if(partnersKey == null) {
				requestPublicKey();
				printChat("Asking for key.\n");
				return;
		}
		printChat("Sending shared secret.\n");

		sharedKey = new byte[160];
		random.nextBytes(sharedKey);
		msg =  new String(sharedKey);
		/*
		System.out.println("MSG:=" + msg);
		System.out.println("MSG:=" + rsa.sign(new java.math.BigInteger(msg.getBytes())));
		System.out.println("MSG:=" + rsa.decrypt(new java.math.BigInteger(msg.getBytes())));
		*/
		msg = "\3" + rsa.encrypt(new java.math.BigInteger(msg.getBytes()), partnersKey);
		wt.getWTalkProtocol().sendMessage(CurrentRoom, msg );
	}

}
